/*
This file is part of Adapa (Automatic DAta PArallelism)
Copyright (C) 2009  Pedro Ribeiro (CRACS & INESC-Porto LA, U. Porto)

Adapa is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 3, as
published by the Free Software Foundation.

Adapa is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Adapa.  If not, see <http://www.gnu.org/licenses/>.
*/
 
#include "Action.h"
#include "Common.h"
#include "Condor.h"
#include "Error.h"
#include "Division.h"

#include <iostream>

// --------------------------------------

/* Basic constructor - set Parameters pointer and create Condor object*/
Action::Action(Parameters *myp) {
  p = myp;

  e = new Condor();
  e->setParameters(p);
}

// --------------------------------------

/* Decide which action to take based on the argument "ACTION" */
void Action::decideAction() {
  if (p->arg["ACTION"]=="STATUS") {

    if (p->arg["STATE"]=="SUBMITTED")
      status();
    
    if (p->arg["STATE"]=="NOTHING")	      
      return;

  } else if (p->arg["ACTION"]=="PREPARE") {

    p->enableParameters();

    divideData();
    submissionFiles();    
    p->arg["STATE"] = "PREPARED";

  } else if (p->arg["ACTION"]=="SUBMIT") {
    
    p->enableParameters();  
    submitJobs();
    p->arg["STATE"] = "SUBMITTED";
        
  } else if (p->arg["ACTION"]=="COLLECT") {
    if (p->arg["STATE"]!="COMPUTED")
      Error::output("s", "Output can only be aggregated after all jobs finished running!");
    p->enableParameters();  
    collect();
    p->arg["STATE"] = "COMPLETED";
    
  } else if (p->arg["ACTION"]=="ALL") {
    p->enableParameters();  
    createExperimentDir();
    divideData();
    submissionFiles();
    submitJobs();
    p->arg["STATE"] = "SUBMITTED"; 
    puts(SEP_AS);
    printf("NOTE: \"ALL\" action not completely supported.\n"
	   "      Jobs already submitted but you must check manually\n"
	   "      for job completion, using adapa with \"STATUS\" action\n");
    puts(SEP_AS);
    
  } else if (p->arg["ACTION"]=="CANCEL") {
    if (p->arg["STATE"]!="SUBMITTED")
      Error::output("s", "Jobs can be cancelled only when they are running!");
    cancelJobs();
    p->arg["STATE"] = "PREPARED";
    
  } else if (p->arg["ACTION"]=="RERUN") {
    Error::output("s","RERUN action not supported in this version");

  } else if (p->arg["ACTION"]=="REMOVE") {
    if (p->arg["STATE"]=="NOTHING")
      Error::output("s", "You can only remove an experiment that has already started!");

    if (Common::yesNoQuestion("Are you sure you want to completely remove experiment?")) {
      if (p->arg["STATE"]=="SUBMITTED")
	cancelJobs();
      p->arg["STATE"] = "PREPARED";
      removeExperimentDir();
    }
    return;

  } else {
    Error::parameterInvalid(*p, "ACTION"); 
  }

  p->writeExperimentConfigFile();
}

// --------------------------------------

/* Divide input data */
void Action::divideData() {
  Division d(p);  

  if (p->arg["STATE"]!="NOTHING")
    if (!Common::yesNoQuestion("Data appears to be already divided. Divide again?"))
      exit(0);
    
  createExperimentDir();
  d.createPartitions();
  d.createPartitionFiles();
}

// --------------------------------------

/* Create submission files */
void Action::submissionFiles() {
  int aux;
  char cmd[MAX_BUF];

  if (p->arg["EXECUTABLE"].empty())
    Error::parameterInvalid(*p, "EXECUTABLE");

  printf("Copying executable file\n");
  sprintf(cmd, "cp -f \"%s\" %s/%s",
	  p->arg["EXECUTABLE"].c_str(),
	  p->arg["DIR"].c_str(),
	  p->arg["EXECUTABLE_COPY"].c_str());
  printf("Running \"%s\"\n", cmd);
  if ((aux=system(cmd))) exit(aux);
  
  e->prepareFiles();
}

// --------------------------------------

/* Check workflow status */
void Action::status() {
  int count;

  e->status();

  count = e->jobsRunning();
  
  if (count==0) {
    printf("No more jobs running! (waiting for aggregation)\n");
    p->arg["STATE"]="COMPUTED";
  } else {
    printf("%d Job%s running\n", count, count==1?"":"s");
  }
}

// --------------------------------------

/* Submit jobs */
void Action::submitJobs() {
  if (p->arg["STATE"]=="NOTHING")
    Error::output("s", "Submission is only possible when data is already divided!");

  if (p->arg["STATE"]!="PREPARED")
    if (!Common::yesNoQuestion("It appears that the jobs have already been submitted.\n Submit again?"))      
      exit(0);
    else
      cancelJobs();
  
  e->setParameters(p);
  e->submit();
}

// --------------------------------------

/* Collect results */
void Action::collect() {
  int i, aux;

  Division d(p);
  d.createPartitionFilesNames();

  printf("Collecting results\n");

  string cmd = "cd " + p->arg["DIR"] + " && " +  "cat";
  for (i=0; i<p->nr_partitions; i++)
    cmd += " " + p->part_output[i];
  cmd += " > " + p->arg["GLOBAL_OUTPUT"];

  printf("Running \"%s\"\n", cmd.c_str());
  if ((aux=system(cmd.c_str()))) exit(aux);

  printf("Done. Experiment complete.\nCheck results of computation in %s/%s\n",
	 p->arg["DIR"].c_str(), p->arg["GLOBAL_OUTPUT"].c_str());
}

// --------------------------------------

/* Cancel all running jobs*/
void Action::cancelJobs() {
  e->setParameters(p);
  e->cancel();
}

// --------------------------------------

/* Create experiment directory */
void Action::createExperimentDir() {
  int aux;
  char cmd[MAX_BUF];
  sprintf(cmd, "mkdir -v -p %s", p->arg["DIR"].c_str());
  printf("Running \"%s\"\n", cmd);
  if ((aux=system(cmd))) exit(aux);
}

// --------------------------------------

/* Remove experiment directory */
void Action::removeExperimentDir() {
  int aux;
  char cmd[MAX_BUF];
  
  sprintf(cmd, "rm -rf %s/", p->arg["DIR"].c_str());
  printf("Running \"%s\"\n", cmd);
  if ((aux=system(cmd))) exit(aux);
 
}

// --------------------------------------

