/*
This file is part of Adapa (Automatic DAta PArallelism)
Copyright (C) 2009  Pedro Ribeiro (CRACS & INESC-Porto LA, U. Porto)

Adapa is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 3, as
published by the Free Software Foundation.

Adapa is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Adapa.  If not, see <http://www.gnu.org/licenses/>.
*/
 
#include "Condor.h"
#include "Common.h"
#include "Error.h"

#include <stdio.h>

// --------------------------------------

void Condor::setParameters(Parameters *myp) {
  p = myp;
}

// --------------------------------------

/* Prepare submission files */
void Condor::prepareFiles() {
  int i, j;
  FILE *f;
  char buf[MAX_BUF];

  puts("Preparing Condor Submission Files");

  sprintf(buf, "%s/%s",
	  p->arg["DIR"].c_str(),
	  p->arg["CONDOR_JOB_FILE"].c_str());

  if (!(f=fopen(buf,"w")))
    Error::output("sss","Could not create file \"",buf,"\"\n");

  fprintf(f,"Universe       = vanilla\n");
  fprintf(f,"Rank           = LoadAvg\n");

  fprintf(f,"Executable     = %s\n", p->arg["EXECUTABLE_COPY"].c_str());
  if (!p->arg["ARGUMENTS"].empty())
    fprintf(f,"Arguments      = %s\n", p->arg["ARGUMENTS"].c_str());

  if (p->transfer_files) {
    fprintf(f,"should_transfer_files = YES\n");
    fprintf(f,"when_to_transfer_output = ON_EXIT\n");
  }

  for (i=0; i<p->nr_partitions; i++) {
    fprintf(f,"\nInput    = %s\n", p->part_input[i].c_str());
    fprintf(f,"Output   = %s\n", p->part_output[i].c_str());
    fprintf(f,"Log      = %s\n", p->part_log[i].c_str());  
    
    if (p->transfer_files) {
      fprintf(f,"transfer_input_files = ");
      for (j=0; j<p->nr_channels; j++)
	fprintf(f,"%s%s",j?",":"",p->part_channel[i][j].c_str());
      fputc('\n',f);
    }
    fprintf(f,"\nQueue\n");
  }

  fclose(f);

  puts(SEP);

}

// --------------------------------------

/* Submit jobs */
void Condor::submit() {
  char *myp, cmd[MAX_BUF], buf[MAX_BUF];
  int aux, cluster=-1;
  FILE *f;

  puts("Submitting to Condor");

  sprintf(cmd, p->arg["CONDOR_SUBMIT_CMD"].c_str(), p->arg["CONDOR_JOB_FILE"].c_str());
  sprintf(buf, "cd %s && %s", p->arg["DIR"].c_str(), cmd);

  printf("Running \"%s\"\n", cmd);
  if (!(f = popen(buf, "r")))
    Error::output("s","Could not submit!\n");
  while (fgets(buf, MAX_BUF, f)) {
    printf("%s", buf);
    if ((myp=strstr(buf, "cluster")))
      sscanf(myp,"cluster %d", &cluster);
  }
  if ((aux=pclose(f))) exit(aux);
  
  if (cluster>=0) {   
    sprintf(buf, "%d", cluster);
    p->arg["CONDOR_CLUSTER"]=buf;
  }
  else
    printf("Could not determine Condor cluster!\n");
}

// --------------------------------------

/* Check jobs status */
void Condor::status() {
  int aux;
  char cmd[MAX_BUF];

  printf("Condor Jobs Status:\n");

  if (p->arg["CONDOR_CLUSTER"].empty())
    printf("Could not determine Condor cluster! (showing all clusters)\n");

  sprintf(cmd, "condor_q %s", p->arg["CONDOR_CLUSTER"].c_str());
  printf("Running \"%s\"\n", cmd);
  if ((aux=system(cmd))) exit(aux);
}

// --------------------------------------

/* Count number of jobs running */
int Condor::jobsRunning() {
  int count, aux;
  char cmd[MAX_BUF], buf[MAX_BUF];
  FILE *f;
  
  count=0;

  if (p->arg["CONDOR_CLUSTER"].empty())
    printf("Could not determine Condor cluster! (checking all clusters)\n");
  
  sprintf(cmd, "condor_q %s -format \"%%s\\n\" Owner",
	  p->arg["CONDOR_CLUSTER"].c_str());
  
  printf("Running \"%s\"\n", cmd);
  if ((f = popen(cmd, "r"))) {
    while (fgets(buf, MAX_BUF, f))
      count++;
    if ((aux=pclose(f))) exit(aux);
  }
  
  return count;  
}

// --------------------------------------

/* Cancel all running jobs */
void Condor::cancel() {
  int aux;
  char cmd[MAX_BUF];

  printf("Removing Condor Jobs:\n");

  if (p->arg["CONDOR_CLUSTER"].empty()) {
    printf("Could not determine Condor cluster for removing! (showing all clusters)\n");
    sprintf(cmd, "condor_q");
    printf("Running \"%s\"\n", cmd);
    if ((aux=system(cmd))) exit(aux);
  } else {
    sprintf(cmd, "condor_rm %s", p->arg["CONDOR_CLUSTER"].c_str());
    printf("Running \"%s\"\n", cmd);
    if ((aux=system(cmd))) 
      printf("Jobs not removed!!\n");
  }
    
}

// --------------------------------------

void Condor::rerun(int i) {
}

