/*
This file is part of Adapa (Automatic DAta PArallelism)
Copyright (C) 2009  Pedro Ribeiro (CRACS & INESC-Porto LA, U. Porto)

Adapa is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 3, as
published by the Free Software Foundation.

Adapa is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Adapa.  If not, see <http://www.gnu.org/licenses/>.
*/
 
/*
Adapa (Automatic DAta PArallelism)
Copyright (C) 2009  Pedro Ribeiro (CRACS & INESC-Porto LA, U. Porto)

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 3, as
published by the Free Software Foundation.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include "Division.h"
#include "Error.h"
#include "Common.h"
#include "Clock.h"

#include <stdio.h>
#include <stdlib.h>

// --------------------------------------

Division::Division(Parameters *aux) {
  p = aux;
}

// --------------------------------------

/* Initialize partition variables */
void Division::initPartitions(int nr) {
  p->nr_partitions = nr; 
  p->start = new int[nr];
  p->end   = new int[nr];

  char buf[MAX_BUF];
  sprintf(buf, "%d", nr);
  p->arg["NUM_PARTITIONS"] = buf;
}

// --------------------------------------

/* Show current partitions */
void Division::showPartitions() {
  int i;
  printf("%d partitions:\n", p->nr_partitions);
  for (i=0; i<p->nr_partitions; i++)
    printf("  - [%d, %d]\t size=%d\n", p->start[i], p->end[i], p->end[i]-p->start[i]+1);
}

// --------------------------------------

/* Create equal sized partitions */
void Division::createEqualPartitions(int n) {
    initPartitions(n);
      
    int part_size = p->nr_values/n;
    p->start[0]=0;
    for (int i=0; i+1<n; i++) {
      p->end[i] = p->start[i]+part_size-1;
      if ((p->nr_values-p->end[i]-1) % (part_size)) p->end[i]++;
      p->start[i+1]=p->end[i]+1;
    }
    p->end[n-1]=p->nr_values-1;  
}

// --------------------------------------

/* Create sliding window partitions */
void Division::createSlidingPartitions(int size, int slide, int num) {
  int i, pos, left;

  int nr_windows = (p->nr_values - size) / slide + 1;
  if (num > nr_windows) num = nr_windows;
  int nr_job = nr_windows / num;

  initPartitions(num);

  for (i=0, pos=0, left=nr_windows; i<num; i++) {
    p->start[i] = pos;
    p->end[i]   = pos + (nr_job-1) * slide + size - 1;
    left -= nr_job;
    if (nr_job * (num-i-1) < left) {
      p->end[i] += slide;
      left--;
    }
    if (p->end[i] > p->nr_values) {
      p->end[i] = p->start[i] + size - 1;
      while (p->end[i]+slide < p->nr_values)
	p->end[i] += slide;
    }
    pos = p->end[i]-size+1+slide;
  }
}

// --------------------------------------

/* Create custom defined partitions */
void Division::createCustomPartitions(int nr) {
  int i;
  char buf[MAX_BUF];
  
  initPartitions(nr);
  for (i=0; i<nr; i++) {
    sprintf(buf, "START_%d", i+1);
    if (sscanf(p->arg[buf].c_str(), "%d", &(p->start[i]))!=1 ||
	p->start[i]<0 || p->start[i]>=p->nr_values)
      Error::parameterInvalid(*p, buf);    
    sprintf(buf, "END_%d", i+1);
    if (sscanf(p->arg[buf].c_str(), "%d", &(p->end[i]))!=1 ||
	p->end[i]<0 || p->end[i]>=p->nr_values)
      Error::parameterInvalid(*p, buf);    
  }
}

// --------------------------------------

/* Create partitions (general function) */
void Division::createPartitions() {
  int n;

  puts("Creating partitions");
  
  if (p->arg["PARTITION_TYPE"]=="EQUAL_SIZED_NUMBER") {
    if (sscanf(p->arg["EQUAL_SIZE_VALUE"].c_str(), "%d", &n)!=1 || n<1)
      Error::parameterInvalid(*p, "EQUAL_SIZE_VALUE");
    printf("Method: Equal sized %d partitions\n", n);
    printf("Size of each partition ~= %d\n", p->nr_values/n);    
    createEqualPartitions(n);

  } else if (p->arg["PARTITION_TYPE"]=="EQUAL_SIZED_SIZE") {
    if (sscanf(p->arg["EQUAL_SIZE_VALUE"].c_str(), "%d", &n)!=1 || n<1)
      Error::parameterInvalid(*p, "EQUAL_SIZE_VALUE");
    printf("Method: Equal sized partitions, each with %d values\n", n);
    printf("Number of partitions ~= %d\n", p->nr_values/n);    
    createEqualPartitions(p->nr_values/n);

  } else if (p->arg["PARTITION_TYPE"]=="SLIDING_WINDOW") {
    int size, slide;
    if (sscanf(p->arg["WINDOW_SIZE"].c_str(), "%d", &n)!=1 || n<1)
      Error::parameterInvalid(*p, "WINDOW_SIZE");
    size = n;
    if (sscanf(p->arg["WINDOW_SLIDE"].c_str(), "%d", &n)!=1 || n<1)
      Error::parameterInvalid(*p, "WINDOW_SLIDE");
    slide=n;
    if (sscanf(p->arg["WINDOW_PARTITIONS"].c_str(), "%d", &n)!=1 || n<1)
      Error::parameterInvalid(*p, "WINDOW_PARTITIONS");
    createSlidingPartitions(size, slide, n);
    
  } else if (p->arg["PARTITION_TYPE"]=="CUSTOM") {
    if (sscanf(p->arg["CUSTOM_NUMBER"].c_str(), "%d", &n)!=1 || n<1)
      Error::parameterInvalid(*p, "CUSTOM_NUMBER");
    printf("Method: %d custom partitions\n", n);
    createCustomPartitions(n);
  } else {
    Error::parameterInvalid(*p, "PARTITION_TYPE");
  }

  if (p->verbose) showPartitions();
  puts(SEP);

}

// --------------------------------------

/* Create partition file names */
void Division::createPartitionFilesNames() {
  int i, j;
  char name[MAX_BUF];

  p->part_input   = new string  [p->nr_partitions];
  p->part_output  = new string  [p->nr_partitions];
  p->part_log      = new string  [p->nr_partitions];
  p->part_channel = new string *[p->nr_partitions];
  for (i=0; i<p->nr_partitions; i++)
    p->part_channel[i] = new string[p->nr_channels];
  
  for (i=0; i<p->nr_partitions; i++) {
    sprintf(name,p->arg["PART_INPUT"].c_str(), i+1);
    p->part_input[i] = string(name);
    sprintf(name,p->arg["PART_OUTPUT"].c_str(), i+1);
    p->part_output[i] = string(name);
    sprintf(name,p->arg["PART_LOG"].c_str(), i+1);
    p->part_log[i] = string(name);
    for (j=0; j<p->nr_channels; j++) {
      sprintf(name,p->arg["PART_DATA"].c_str(), i+1, j+1);
      p->part_channel[i][j] = string(name);
    }
  }
  
}

// --------------------------------------

/* Create partitition data files */
void Division::createPartitionFiles() {

  puts("Creating Partition Files");
  createPartitionFilesNames();
  createPartitionFilesInput();

  Clock c;
  c.startTimer();

  if (p->arg["DATA_TYPE"]=="ASCII")
    createPartitionFilesASCII();
  else if (p->arg["DATA_TYPE"]=="BINARY")
    createPartitionFilesBinary();
  else
    Error::parameterInvalid(*p, "DATA_TYPE");    

  c.endTimer();
  printf("Real Time spent on partition files: %.2lfs\n", c.secondsElapsed());

  puts(SEP);
}

// --------------------------------------

/* Create partition data files when they are in binary format */
void Division::createPartitionFilesBinary() {
  int i, j, left, aux, b_start, b_end;
  char buf[MAX_BUF], buf_io[BUFSIZ];
  FILE *out, *in;

  for (i=0; i<p->nr_channels; i++) {

    if (!(in=fopen(p->channel[i].c_str(),"r")))
      Error::output("sss","Could not open file \"",p->channel[i].c_str(),"\"\n");


    for (j=0; j<p->nr_partitions; j++) {
      sprintf(buf, "%s/%s", p->arg["DIR"].c_str(), p->part_channel[j][i].c_str());
      if (!(out=fopen(buf,"w")))
	Error::output("sss","Could not create file \"",buf,"\"\n");    
    
      b_start = p->start[j] * p->bytes_per_value;
      b_end   = p->end[j]   * p->bytes_per_value + p->bytes_per_value;
      left = b_end - b_start;
      fseek(in, b_start, SEEK_SET);
      while (left) {
	aux = BUFSIZ;
	if (left<aux) aux=left;
	fread(buf_io, 1, aux, in);
	fwrite(buf_io, 1, aux, out);
	left-=aux;
      }
      fclose(out);

      putchar('.'); fflush(stdout);    
    }
    
    fclose(in);
  }
  putchar('\n'); fflush(stdout);
}

// --------------------------------------

/* Create partition data files when they are in ASCII format */
void Division::createPartitionFilesASCII() {
  int i, j, count;
  char buf[MAX_BUF];
  FILE *f[p->nr_partitions], *org;

  for (i=0; i<p->nr_channels; i++) {
    for (j=0; j<p->nr_partitions; j++) {
      sprintf(buf, "%s/%s", p->arg["DIR"].c_str(), p->part_channel[j][i].c_str());
      if (!(f[j]=fopen(buf,"w")))
	Error::output("sss","Could not create file \"",buf,"\"\n");
    }
   
    if (!(org=fopen(p->channel[i].c_str(),"r")))
      Error::output("sss","Could not open file \"",p->channel[i].c_str(),"\"\n");
    count=0;
    while (fgets(buf,MAX_BUF,org)) {
      for (j=0; j<p->nr_partitions; j++) {
	if (count>=p->start[j] && count <=p->end[j])
	  fputs(buf, f[j]);
      }
      count++;
      if (count%DOT_WRITE==0) {putchar('.'); fflush(stdout);}
    }
    for (j=0; j<p->nr_partitions; j++)
      fclose(f[j]);
    
    fclose(org);
    putchar('.'); fflush(stdout);    
  }
  putchar('\n'); fflush(stdout);
}

// --------------------------------------

/* Create input files for each job */
void Division::createPartitionFilesInput() {
  int i, j;
  char buf[MAX_BUF];

  FILE *input;
  for (i=0; i<p->nr_partitions; i++) {
    sprintf(buf, "%s/%s", p->arg["DIR"].c_str(), p->part_input[i].c_str());
    if (!(input=fopen(buf,"w")))
      Error::output("sss","Could not create file \"",buf,"\"\n");
    fprintf(input, "%d\n%d\n%d\n", i+1,
	    (p->end[i])-(p->start[i])+1, p->nr_channels);
    for (j=0; j<p->nr_channels; j++) {
      fprintf(input, "%s\n", p->part_channel[i][j].c_str());
    }
  }
}
