/*
This file is part of Adapa (Automatic DAta PArallelism)
Copyright (C) 2009  Pedro Ribeiro (CRACS & INESC-Porto LA, U. Porto)

Adapa is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 3, as
published by the Free Software Foundation.

Adapa is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Adapa.  If not, see <http://www.gnu.org/licenses/>.
*/
 
#include "Parameters.h"
#include "Error.h"
#include "Common.h"

#include <stdio.h>
#include <ctype.h>

// --------------------------------------

/* Show help */
void Parameters::showUsage() {
  printf("USAGE:\n\n"
	 "adapa <experiment directory> <action> [<any other options>]\n\n"
	 "<action> can be:\n"
	 "  - \"status\"  Report current workflow state\n"
	 "  - \"prepare\" Divide input data and create submission files\n"
	 "  - \"submit\"  Submit jobs to submission engine\n"
	 "  - \"collect\" Aggregate results obtained from submission\n"
	 "  - \"all\"     Prepare, submit and collect results\n"
	 "  - \"cancel\"  Stop any jobs that are running\n"
	 "  - \"remove\"  Cancel and remove the entire experiment\n\n"
	 "Other options come in the form --PARAMETER VALUE (see manual for details)\n\n"
	 );
}

// --------------------------------------

/* Show current workflow state */
void Parameters::msgState() {
  puts(SEP_AS);
  printf("Experiment directory: \"%s\"\n", arg["DIR"].c_str());
  printf("Workflow State: ");
  if (arg["STATE"]=="NOTHING")
    puts("Experiment still not started\n(waiting for an action)");
  else if (arg["STATE"]=="PREPARED")
    puts("Data division done and submission files created\n"
	 "(waiting for submission order)");
  else if (arg["STATE"]=="SUBMITTED")
    puts("Jobs submitted\n(waiting for completion)");
  else if (arg["STATE"]=="COMPUTED")
    puts("All jobs already computed\n(waiting for aggregation)");
  else if (arg["STATE"]=="COMPLETED")
    puts("Experiment seems to be complete");
  else {
    puts("Unknown experiment state (ERROR)!!");
    exit(1);
  }
  puts(SEP_AS);
}

// --------------------------------------

/* Initialize default parameters */
void Parameters::initDefaultParameters() {
  arg["ACTION"] = "STATUS";
  arg["STATE"]  = "NOTHING";

  arg["DIR"]        = "./experiment";
  arg["DATA_TYPE"]       = "ASCII";

  arg["COPY_BUFFER"]     = "8192";
  arg["PART_DATA"]       = "data-p%d-c%d";
  arg["PART_INPUT"]      = "input%d.txt";
  arg["PART_OUTPUT"]     = "output%d.txt";
  arg["PART_LOG"]        = "log%d.txt";

  arg["GLOBAL_OUTPUT"]   = "output.txt";

  arg["CONDOR_JOB_FILE"] = "condor.job";
  arg["CONDOR_SUBMIT_CMD"] = "condor_submit %s";
}

// --------------------------------------

/* Parse command line arguments */
void Parameters::parseParameters(int argc, char **argv) {
  int i;

  for (i=1; i<argc; i++) {
    if (argv[i][0]=='-' && argv[i][1]=='-' && (i+1)<argc) {
      arg[&argv[i][2]] = argv[i+1];      
      i++;
    }
  }  
}

// --------------------------------------

/* Write a configuration file */
void Parameters::writeConfigFile(const char *s) {
  FILE *f;
  f = fopen(s,"w");
  if (!f)
    Error::output("sss","Could not write configuration file \"",s, "\"");

  if (verbose) printf("Writing configuration file \"%s\"\n",s);

  fputs("# Adapa configuration file\n", f);
  fputs("# (automatically generated)\n", f);
  fputs("# -------------------------\n", f);

  map<string, string>::iterator ii;
  for(ii=arg.begin(); ii!=arg.end(); ++ii)
    if ((string)((*ii).first) != "ACTION" && 
	(string)((*ii).first) != "CONFIG_FILE")
      fprintf(f,"%s=%s\n",(*ii).first.c_str(), (*ii).second.c_str());

  fclose(f);
}

// --------------------------------------

/* Read config file on experiment directory */
void Parameters::readExperimentConfigFile() {
  char buf[MAX_BUF];
  FILE *f;
  sprintf(buf, "%s/%s", arg["DIR"].c_str(), CONFIG_FILE);
  if ((f=fopen(buf, "r"))) {
    fclose(f);
    readConfigFile(buf);
  }
}

// --------------------------------------

/* Write config file on experiment directory */
void Parameters::writeExperimentConfigFile() {
  char buf[MAX_BUF];
  sprintf(buf, "%s/%s", arg["DIR"].c_str(), CONFIG_FILE);
  writeConfigFile(buf);
}

// --------------------------------------

/* Read configuration file given on command line */
void Parameters::readArgumentConfigFile() {
  if (!arg["CONFIG_FILE"].empty())
    readConfigFile(arg["CONFIG_FILE"].c_str());
}

// --------------------------------------

/* Read a given configuration file */
void Parameters::readConfigFile(const char *s) {
  int i, pos;
  char par[MAX_BUF], value[MAX_BUF], buf[MAX_BUF];
  FILE *f;

  f = fopen(s,"r");
  if (!f)
    Error::output("sss","Could not read configuration file \"", s, "\"");

  if (verbose) printf("Reading configuration file \"%s\"\n", s);

  while (fgets(buf, MAX_BUF, f)) {
    par[0]=value[0]=0;
    if (buf[0]==COMMENT) continue;
    for (pos=0; isspace(buf[pos]); pos++);
    if (!buf[pos]) continue;
    for (i=0; !isspace(buf[pos]) && buf[pos]!='='; pos++, i++)
      par[i] = buf[pos];
    par[i]=0;
    for (; isspace(buf[pos]) || buf[pos]=='='; pos++);
    if (!buf[pos]) continue;
    for (i=0; buf[pos]!='\n' && buf[pos]!='\r'; pos++, i++)
      value[i] = buf[pos];
    value[i]=0;
    
    if (par[0] && value[0]) {	
      arg[par]=value;
    }
  }
    
  fclose(f);
}

// --------------------------------------

/* Enable and check parameter values */
void Parameters::enableValues() {
  int n;
  if (sscanf(arg["NUM_CHANNELS"].c_str(), "%d", &n)!=1 || n<1)
    Error::parameterInvalid(*this, "NUM_CHANNELS");
  printf("Number of channels: %d\n", n);
  nr_channels = n;

  if (sscanf(arg["NUM_VALUES"].c_str(), "%d", &n)!=1 || n<1)
    Error::parameterInvalid(*this, "NUM_VALUES");
  printf("Number of values per channel: %d\n", n);
  nr_values = n;

  if (!arg["NUM_PARTITIONS"].empty() &&
      sscanf(arg["NUM_PARTITIONS"].c_str(), "%d", &n)!=1 || n<1)
    Error::parameterInvalid(*this, "NUM_PARTITIONS");
  nr_partitions = n;

  if (arg["DATA_TYPE"] != "ASCII" &&
      arg["DATA_TYPE"] != "BINARY")    
    Error::parameterInvalid(*this, "DATA_TYPE");
  printf("Data type: %s\n", arg["DATA_TYPE"].c_str());

  if (arg["DATA_TYPE"] == "BINARY" &&
      (sscanf(arg["BYTES_PER_VALUE"].c_str(), "%d", &n)!=1 || n<1))
    Error::parameterInvalid(*this, "BYTES_PER_VALUE");
  bytes_per_value = n;

  if (arg["TYPE_CHANNELS"] == "SINGLE") {
    Error::output("s", "Single file with multiple channels not supported yet!");

  } else if (arg["TYPE_CHANNELS"] == "MULTIPLE") {
    puts("Type of channels: multiple files (one channel per file)");
    channel = new string[nr_channels];

    if (arg["MULTIPLE_NAME"] == "CUSTOM") {
      char name[MAX_BUF], aux[MAX_BUF];      
      for (int i=1; i<=nr_channels; i++) {
	sprintf(aux, "MULTIPLE_CUSTOM_%d", i);
	if (sscanf(arg[aux].c_str(), "%s", name)!=1)
	  Error::parameterInvalid(*this, aux);
	channel[i-1] = string(name);
      }
    } else {      
      char name[MAX_BUF], aux[MAX_BUF];
      if (sscanf(arg["MULTIPLE_NAME"].c_str(), "%s", name)!=1)
	Error::parameterInvalid(*this, "MULTIPLE_NAME");
      for (int i=1; i<=nr_channels; i++) {
	sprintf(aux, name, i);
	channel[i-1] = string(aux);      
      }
    }    

    if (verbose) {
      printf("%d channel files:\n", nr_channels);
      for (int i=0; i<nr_channels; i++)
	printf("   - Channel %d: %s\n", i+1, channel[i].c_str());
    }
  }

  if (arg["EXECUTABLE_COPY"].empty()) {
    char buf[MAX_BUF];
    Common::fileName(arg["EXECUTABLE"].c_str(), buf);
    arg["EXECUTABLE_COPY"] = buf;
  }

  puts(SEP);
}

// --------------------------------------

/* Enable and check parameter values */
void Parameters::enableParameters() {

  if (arg["VERBOSE"]=="TRUE") verbose=true; 
  else verbose = false;

  if (arg["TRANSFER_FILES"]=="TRUE") transfer_files=true; 
  else transfer_files = false;

  enableValues();
}

// --------------------------------------

/* Do all parameter related stuff */
void Parameters::doParameters(int argc, char **argv) {
  int i;
  char buf[MAX_BUF];

  initDefaultParameters();

  if (argc < 3 ||
      (!strcmp(argv[2], "status") &&
       !strcmp(argv[2], "prepare") &&
       !strcmp(argv[2], "submit") &&
       !strcmp(argv[2], "collect") &&
       !strcmp(argv[2], "all") &&
       !strcmp(argv[2], "cancel") &&
       !strcmp(argv[2], "remove"))) {
    showUsage();
    exit(1);
  }

  arg["DIR"] = argv[1];
  for (i=0; argv[2][i]; i++)
    buf[i] = toupper(argv[2][i]);
  buf[i]=0;
  arg["ACTION"] = buf;

  parseParameters(argc, argv);
  readArgumentConfigFile();
  if (arg["VERBOSE"]=="TRUE") verbose=true; 
  else verbose = false;
  
  readExperimentConfigFile();
  readArgumentConfigFile();
  parseParameters(argc, argv);
}
